// Copyright 2021 4inka

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:

// 1. Redistributions of source code must retain the above copyright notice,
// this list of conditions and the following disclaimer.

// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.

// 3. Neither the name of the copyright holder nor the names of its contributors
// may be used to endorse or promote products derived from this software without
// specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
// NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
// WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.

library easy_pin_code;

import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

enum EasyPinCodeType {
  filledSquare,
  filledCircle,
  displayedValueSquare,
  displayedValueCircle
}

class EasyPinCode extends StatefulWidget {
  /// Can be used to set the 
  final int length;
  final double fieldWidth;
  final double fieldHeight;
  final bool autoFocus;
  final bool obscureText;
  final Function(String value) onSubmit;
  final List<TextInputFormatter> inputFormatters;
  final TextInputType keyboardType;
  final String obscuringCharacter;
  final TextEditingController? controller;
  final Duration animationDuration;
  final EasyPinCodeType type;
  final double marginBetween;
  final bool unfocusOnSubmit;
  final bool autoSubmit;


  final TextStyle textStyle;
  final Color unfocusedBorderColor;
  final Color focusedBorderColor;
  final Color filledBorderColor;
  /// Used when waiting on `onSubmit` method to finish
  final Color disabledBorderColor;

  final Color unfocusedBackgroundColor;
  final Color focusedBackgroundColor;
  final Color filledBackgroundColor;
  /// Used when waiting on `onSubmit` method to finish
  final Color disabledBackgroundColor;

  final Color cursorColor;

  final double borderWidth;

  const EasyPinCode({
    Key? key,
    required this.length,
    required this.onSubmit,
    this.autoSubmit = false,
    this.fieldWidth = 40,
    this.fieldHeight = 40,
    this.autoFocus = true,
    this.obscureText = false,
    this.obscuringCharacter = '●',
    this.controller,
    this.animationDuration = const Duration(milliseconds: 150),
    this.type = EasyPinCodeType.displayedValueSquare,
    this.marginBetween = 5,
    this.unfocusOnSubmit = false,
    this.inputFormatters = const [],
    this.keyboardType = TextInputType.visiblePassword,



    this.textStyle = const TextStyle(
      color: Colors.black,
    ),
    this.unfocusedBorderColor = Colors.grey,
    this.focusedBorderColor = Colors.blueAccent,
    this.filledBorderColor = Colors.blue,
    this.disabledBorderColor = Colors.grey,
    this.unfocusedBackgroundColor = Colors.transparent,
    this.focusedBackgroundColor = Colors.blueAccent,
    this.filledBackgroundColor = Colors.blue,
    this.disabledBackgroundColor = Colors.grey,
    this.cursorColor = Colors.blue,
    this.borderWidth = 2
  }) : assert(length >= 2, 'length cannot be less than 2'),
    assert(obscuringCharacter.length == 1, 'obscuringCharacter cannot be empty or have more than one character'),
    super(key: key);

  @override
  _EasyPinCodeState createState() => _EasyPinCodeState();
}

class _EasyPinCodeState extends State<EasyPinCode> with TickerProviderStateMixin {
  late TextEditingController _textEditingController;
  late FocusNode _focusNode;
  final List<AnimationController> _animationControllers = [];
  final List<Animation<double>> _animations = [];
  final List<AnimationController> _cursorAnimationControllers = [];
  int _currentEditingIndex = 0;
  bool _isReadOnly = false;

  @override
  void initState() {
    super.initState();
    _textEditingController = widget.controller ?? TextEditingController();
    _focusNode = FocusNode();

    for (var i = 0; i < widget.length; i++) {
      _cursorAnimationControllers.add(
        AnimationController(
          duration: const Duration(milliseconds: 400),
          vsync: this,
        )
      );
      _cursorAnimationControllers[i].repeat(reverse: true);
      _animationControllers.add(
        AnimationController(
          duration: widget.animationDuration,
          vsync: this,
        )
      );
      _animations.add(
        CurvedAnimation(
          parent: _animationControllers[i],
          curve: Curves.easeIn
        )
      );
    }

    _textEditingController.addListener(() async {
      if (_textEditingController.text.isNotEmpty || _currentEditingIndex != 0) {
        if (_currentEditingIndex < _textEditingController.text.length) {
          for (var i = 0; i < _textEditingController.text.length; i++) {
            _animationControllers[i].forward();
          }
        }
        else if (_currentEditingIndex > _textEditingController.text.length) {
          for (var i = widget.length - 1; i > _textEditingController.text.length - 1; i--) {
            _animationControllers[i].reverse();
          }
        }
      }

      setState(() => _currentEditingIndex = _textEditingController.text.length);

      if (widget.length == _textEditingController.text.length && widget.autoSubmit) {
        await _executeOnSubmitTask();
      }
    });
  }

  Future<void> _executeOnSubmitTask() async {
    setState(() => _isReadOnly = true );
    if (widget.unfocusOnSubmit) {
      _focusNode.unfocus();
    }
    await widget.onSubmit(_textEditingController.text);
    setState(() => _isReadOnly = false );
  }

  Color _borderColor(int index) {
    if (_currentEditingIndex == index && _focusNode.hasFocus) {
      return widget.focusedBorderColor;
    }
    else if (_textEditingController.text.length >= index && !_isReadOnly) {
      return widget.filledBorderColor;
    }
    else if (_textEditingController.text.length < widget.length) {
      return widget.unfocusedBorderColor;
    }

    // If it gets here, it means that the field is in readOnly mode and _isReadOnly is equal to true
    return widget.disabledBorderColor;
  }

  Color _backgroundColor(int index) {
    if (_currentEditingIndex == index && _focusNode.hasFocus) {print('1');
      return widget.focusedBackgroundColor;
    }
    else if (_textEditingController.text.length >= index && !_isReadOnly) {
      return widget.filledBackgroundColor;
    }
    else if (_textEditingController.text.length < widget.length) {print('4');
      return widget.unfocusedBackgroundColor;
    }

    // If it gets here, it means that the field is in readOnly mode and _isReadOnly is equal to true
    return widget.disabledBackgroundColor;
  }

  @override
  Widget build(BuildContext context) {
    return SizedBox(
      width: (widget.fieldWidth + widget.marginBetween * 2) * widget.length,
      child: Stack(
        children: [
          Opacity(
            opacity: 0,
            child: TextField(
              controller: _textEditingController,
              readOnly: _isReadOnly,
              focusNode: _focusNode,
              maxLength: widget.length,
              autofocus: widget.autoFocus,
              keyboardType: widget.keyboardType,
              inputFormatters: widget.inputFormatters,
              decoration: const InputDecoration(
                counter: SizedBox()
              ),
              onSubmitted: (value) async  => _executeOnSubmitTask()
            )
          ),
          GestureDetector(
            onTap: () {
              _focusNode.requestFocus();
              setState(() {});
            },
            child: Row(
              mainAxisSize: MainAxisSize.min,
              mainAxisAlignment: MainAxisAlignment.spaceBetween,
              crossAxisAlignment: CrossAxisAlignment.center,
              children: List.generate(widget.length, (index) {
                List<String> _characters = _textEditingController.text.split('');

                return Stack(
                  alignment: Alignment.center,
                  children: [
                    Visibility(
                      visible: widget.type == EasyPinCodeType.displayedValueSquare || widget.type == EasyPinCodeType.displayedValueCircle,
                      child: FadeTransition(
                        opacity: _animations[index],
                        child: Container(
                          width: widget.fieldWidth,
                          height: widget.fieldHeight,
                          decoration: BoxDecoration(
                            color: _backgroundColor(index),
                            borderRadius: BorderRadius.circular(widget.type == EasyPinCodeType.displayedValueCircle || widget.type == EasyPinCodeType.filledCircle ? 100 : 5)
                          )
                        )
                      )
                    ),
                    Visibility(
                      visible: widget.type == EasyPinCodeType.filledSquare || widget.type == EasyPinCodeType.filledCircle,
                      child: ScaleTransition(
                        scale: _animationControllers[index],
                        alignment: Alignment.center,
                        child: Container(
                          width: widget.fieldWidth - 7,
                          height: widget.fieldHeight - 7,
                          decoration: BoxDecoration(
                            color: _backgroundColor(index),
                            borderRadius: BorderRadius.circular(widget.type == EasyPinCodeType.displayedValueCircle || widget.type == EasyPinCodeType.filledCircle ? 100 : 5)
                          )
                        )
                      )
                    ),
                    AnimatedContainer(
                      width: widget.fieldWidth,
                      height: widget.fieldHeight,
                      margin: EdgeInsets.all(widget.marginBetween),
                      duration: const Duration(milliseconds: 400),
                      decoration: BoxDecoration(
                        color: Colors.transparent,
                        borderRadius: BorderRadius.circular(widget.type == EasyPinCodeType.displayedValueCircle || widget.type == EasyPinCodeType.filledCircle ? 100 : 5),
                        border: Border.all(
                          color: _borderColor(index),
                          width: widget.borderWidth,
                        )
                      )
                    ),
                    Visibility(
                      visible: widget.type == EasyPinCodeType.displayedValueCircle || widget.type == EasyPinCodeType.displayedValueSquare,
                      child: Text(
                        _characters.isNotEmpty && _characters.length - 1 >= index ?
                          widget.obscureText ? widget.obscuringCharacter : _characters[index] : '',
                        style: widget.textStyle
                      ),
                    ),
                    Visibility(
                      visible: (widget.type == EasyPinCodeType.displayedValueCircle || widget.type == EasyPinCodeType.displayedValueSquare) && index == _currentEditingIndex && _focusNode.hasFocus,
                      child: FadeTransition(
                        opacity: _cursorAnimationControllers[index],
                        child: Container(
                          height: (widget.textStyle.fontSize ?? 16) + 5,
                          width: 2,
                          decoration: BoxDecoration(
                            color: widget.cursorColor,
                            borderRadius: BorderRadius.circular(5)
                          )
                        )
                      )
                    )
                  ]
                );
              })
            ),
          )
        ]
      )
    );
  }
}
